-- handles the main parent frame and event handling

local cql = ClassicQuestLog

CQL_SOLID_BACKDROP_COLOR = CreateColor(0.05,0.05,0.05)
CQL_SLIDER_BACKDROP_STYLE = {
	bgFile = "Interface\\Buttons\\UI-SliderBar-Background",
	edgeFile = "Interface\\Buttons\\UI-SliderBar-Border",
	tile = true,
	tileSize = 8,
	edgeSize = 8,
	insets = { left = 3, right = 3, top = 6, bottom = 6 },
}

--[[ events ]]

function cql:PLAYER_LOGIN()
    cql.chrome:Initialize()
    cql.options:Initialize()
    cql.log:Initialize() -- should be called before OnSizeChanged so all HybridScrollFrame buttons created for max height
    -- set min/max height
    if select(4,GetBuildInfo())>=100000 then
        cql:SetResizeBounds(667,300,667,700)
    else
        cql:SetMinResize(667,300)
        cql:SetMaxResize(667,700) -- if this changes, change in XML too
    end
    -- script handlers for main window
    cql:SetScript("OnSizeChanged",cql.OnSizeChanged) -- not defined in XML to give HybridScrollFrame a chance to initialize above
    cql:OnSizeChanged(self:GetWidth(),self:GetHeight()) -- directly call the OnSizeChanged to set log/detail height
    -- BasicFrameTemplate stuff
    cql.CloseButton:SetScript("OnKeyDown",cql.CloseButtonOnKeyDown)
    -- setup objective tracker override
    if not IsAddOnLoaded("Blizzard_ObjectiveTracker") then
        cql:RegisterEvent("ADDON_LOADED")
    else
        cql:HandleObjectiveTracker()
    end
    -- setup hiding our frame when some default ones show since they can't (easily) co-exist with ours
    local hidefunc = function() cql.wasShown = cql:IsVisible() cql:Hide() end
    QuestLogPopupDetailFrame:HookScript("OnShow",hidefunc)
    WorldMapFrame:HookScript("OnShow",hidefunc)
    QuestFrame:HookScript("OnShow",hidefunc)
    cql:SetMode("detail")
    if C_QuestLog.GetSelectedQuest()==0 then -- if user logs in with no quest, try again in half a second
        C_Timer.After(0.5,cql.Update)
    end
    cql:RegisterEvent("QUEST_ACCEPTED")
    -- no longer having a separate binding; if user had old binding defined, move it to default quest log
    if GetBindingKey("CLASSIC_QUEST_LOG_TOGGLE") and not GetBindingKey("TOGGLEQUESTLOG") then
        SetBinding(GetBindingKey("CLASSIC_QUEST_LOG_TOGGLE"),"TOGGLEQUESTLOG",GetCurrentBindingSet())
    end
    -- and update binding
    cql:UpdateBindingOverride()
    -- add slash command to toggle window
    SLASH_CLASSICQUESTLOG1 = "/cql"
    SlashCmdList["CLASSICQUESTLOG"] = cql.ToggleWindow
    -- broker button
	local ldb = LibStub and LibStub:GetLibrary("LibDataBroker-1.1",true)
    if ldb then
        ldb:NewDataObject("ClassicQuestLog",{ type="launcher", icon="Interface\\AddOns\\Classic Quest Log\\cql-button.tga", tooltiptext="Classic Quest Log", OnClick=cql.ToggleWindow })
	end        
    -- finally, minimap button
    ClassicQuestLogMinimapButton:Setup()
end

function cql:PLAYER_REGEN_DISABLED()
    cql.chrome.mapButton:Disable()
end

function cql:PLAYER_REGEN_ENABLED()
    cql.chrome.mapButton:Enable()
end

function cql:QUEST_DETAIL()
    cql:Hide()
end

function cql:ADDON_LOADED(addon)
    if addon=="Blizzard_ObjectiveTracker" then
        cql:HandleObjectiveTracker()
    end
end

function cql:QUEST_ACCEPTED(questID)
    cql.lastAcceptedQuestID = questID
    if cql:IsVisible() then -- generally, quests are not accepted while cql is up
        cql:Update()
    end
end

-- this event fires either in reaction to either ShowLoreButton OnEnter or SetLore call
function cql:LORE_TEXT_UPDATED_CAMPAIGN(campaignID,lore)
    cql:UnregisterEvent("LORE_TEXT_UPDATED_CAMPAIGN") -- can stop listening
    if campaignID then
        cql.lore.campaignHasLore[campaignID] = lore and #lore>0 -- set whether campaign has lore
        if cql.lore:IsVisible() then -- if lore is visible, update the lore
            cql.lore:Update(campaignID,lore)
        else -- otherwise show the lore button if we're waiting to see if it should be shown
            cql.lore:UpdateLoreButtonVisibility()
        end
    end
end

--[[ main window script handlers ]]

-- for events with a defined function, run them; for the rest do a UI update
function cql:OnEvent(event,...)
    if cql[event] then
        cql[event](self,...)
    else
        cql:Update()
    end
end

-- this handler is only watched after cql.log's HybridScrollFrame_CreateButtons, so that enough buttons are
-- made for the maximum height (700 height is defined in xml and should match the MaxResize height above)
function cql:OnSizeChanged(width,height)
    cql.detail:SetHeight(height-93)
    cql.log:SetHeight(height-93)
    -- update log scrollframe for changing height (no need to update data)
    cql.log:HybridScrollFrameUpdate()
end
 
 -- on mouse down of parent window, start moving unless LockWindow checked (of shift held)
 function cql:OnMouseDown()
    if not cql.options:Get("LockWindow") or IsShiftKeyDown() then
        cql.isWindowMoving = true
        self:StartMoving()
    end
 end

 -- on mouse up of parent window, stop moving if it was moving
 function cql:OnMouseUp()
    if self.isWindowMoving then
        self.isWindowMoving = nil
        self:StopMovingOrSizing()
    end
 end

-- when main window shows, hide anything that we're uncomfortable co-existing with and start listening for events
function cql:OnShow()
    QuestFrame:ClearAllPoints()
    QuestInfoSealFrame:ClearAllPoints()
    -- details frame can only be in one frame at a time; hide any frames that could have one
    if WorldMapFrame:IsVisible() and not InCombatLockdown() then
        ToggleWorldMap() -- can't hide in combat, may be unavoidable to co-exist
    end
    if QuestLogPopupDetailFrame:IsVisible() then
        QuestLogPopupDetailFrame:Hide()
    end
    if QuestFrame:IsVisible() then
        -- when the quest frame is up waiting to accept/deline and is abruptly hidden, it can prevent ESC from working
        -- until the window is dismissed normally; this will click the Close button to prevent the behavior
        if QuestFrameCloseButton:IsVisible() and QuestFrameCloseButton:IsEnabled() then
            QuestFrameCloseButton:Click()
        elseif QuestFrameAcceptButton:IsVisible() and QuestFrameAcceptButton:IsEnabled() then
            QuestFrameAcceptButton:Click() -- for automatic quests that appear (upon login), the close button may be disabled; click accept instead
        else
            QuestFrame:Hide()
        end
    end
    cql:RegisterEvent("QUEST_DETAIL")
    cql:RegisterEvent("QUEST_LOG_UPDATE")
    cql:RegisterEvent("QUEST_WATCH_LIST_CHANGED")
    cql:RegisterEvent("GROUP_ROSTER_UPDATE")
    cql:RegisterEvent("PARTY_MEMBER_ENABLE")
    cql:RegisterEvent("PARTY_MEMBER_DISABLE")
    cql:RegisterEvent("QUEST_POI_UPDATE")
    cql:RegisterEvent("QUEST_WATCH_UPDATE")
    --cql:RegisterEvent("QUEST_ACCEPTED")
    cql:RegisterEvent("UNIT_QUEST_LOG_CHANGED")
    cql:RegisterEvent("PLAYER_REGEN_ENABLED")
    cql:RegisterEvent("PLAYER_REGEN_DISABLED")
    PlaySound(SOUNDKIT.IG_QUEST_LOG_OPEN)
    cql:SetMode("detail")
end

-- whem main window hides, stop listening for events
function cql:OnHide()
    cql:UnregisterEvent("QUEST_DETAIL")
    cql:UnregisterEvent("QUEST_LOG_UPDATE")
    cql:UnregisterEvent("QUEST_WATCH_LIST_CHANGED")
    cql:UnregisterEvent("GROUP_ROSTER_UPDATE")
    cql:UnregisterEvent("PARTY_MEMBER_ENABLE")
    cql:UnregisterEvent("PARTY_MEMBER_DISABLE")
    cql:UnregisterEvent("QUEST_POI_UPDATE")
    cql:UnregisterEvent("QUEST_WATCH_UPDATE")
    --cql:UnregisterEvent("QUEST_ACCEPTED")
    cql:UnregisterEvent("UNIT_QUEST_LOG_CHANGED")
    cql:UnregisterEvent("PLAYER_REGEN_ENABLED")
    cql:UnregisterEvent("PLAYER_REGEN_DISABLED")
    PlaySound(SOUNDKIT.IG_QUEST_LOG_CLOSE)
end

-- rather than UISpecialFrames, using the CloseButton (from BasicFrameTemplate) to capture keys to close the window
function cql:CloseButtonOnKeyDown(key)
    if key==GetBindingKey("TOGGLEGAMEMENU") then
        if cql.mode=="lore" or cql.mode=="options" then -- if on lore or options mode, hide them
            cql:SetMode("detail")
            PlaySound(SOUNDKIT.IG_QUEST_LOG_CLOSE)
        else -- otherwise close Classic Quest Log
            cql:Hide()
        end
        self:SetPropagateKeyboardInput(false)
    else
        self:SetPropagateKeyboardInput(true)
    end
end

--[[ default UI behavior changes ]]

local function onBlockHeaderClick(self,block,mouseButton)
    if IsModifiedClick("CHATLINK") and ChatEdit_GetActiveWindow() then
        return -- user was linking quest to chat
    end
    cql.log:Update()
    if mouseButton~="RightButton" then
        if IsModifiedClick("QUESTWATCHTOGGLE") then
            return -- user was untracking a quest
        end
        if InCombatLockdown() then
            return -- in combat, always let map remain in combat (can't ToggleWorldMap then)
        end
        if not cql.options:Get("ShowFromObjectiveTracker") then
            return -- always let map remain if 'Show For Objective Tracker' is disabled
        end
        -- the selected questID should've been selected in the act of clicking an objective.
        -- if the questID is valid, close the map and open this addon's
        if C_QuestLog.IsOnQuest(C_QuestLog.GetSelectedQuest() or 0) then
            ToggleWorldMap()
            cql:Show()
        end
        -- see if the questID is visible in the log; leave if so
        for _,info in pairs(cql.log.quests) do
            if info.questID == block.id then -- found it, we're done
                return
            end
        end
        -- it's not visible in the log, expand the header that contains the questID if it exists
        cql.log:ExpandHeaderContainingQuestID(block.id)
    end
end

-- this will make clicking a quest on the objective tracker summon Classic Quest Log, unless
-- the Show From Objective Tracker option is unchecked; it should be run only once when objective tracker loads
function cql:HandleObjectiveTracker()
    cql:UnregisterEvent("ADDON_LOADED")
    -- hook clicking of quest objective to summon classic quest log instead of world map
    hooksecurefunc(QUEST_TRACKER_MODULE,"OnBlockHeaderClick",onBlockHeaderClick)
    hooksecurefunc(CAMPAIGN_QUEST_TRACKER_MODULE,"OnBlockHeaderClick",onBlockHeaderClick)
end

--[[ window ]]

function cql:ToggleWindow()
    cql:SetShown(not cql:IsVisible())
end

-- this does a full update of the UI
function cql:Update()
    -- first update the detail/lore/options panel, depending which is up
    if cql.mode and cql[cql.mode].Update then
        cql[cql.mode]:Update() -- call the Update for the mode we're in ("detail", "lore" or "options")
    end
    cql.log:Update() -- rebuilds cql.log.quests and updates log display
    cql.chrome:Update() -- updates panel buttons and other chrome bits
    cql.resizeGrip:SetShown(cql.options:Get("ShowResizeGrip"))
end

-- switches between "detail", "lore" and "options" modes; hiding the unused modes and showing the one being set
local modes = { {"detail","lore","options"}, {"lore","detail","options"}, {"options","detail","lore"} }
function cql:SetMode(mode)
    for i=1,#modes do
        if mode==modes[i][1] and cql.mode~=mode then
            cql.mode = mode
            cql[modes[i][2]]:Hide()
            cql[modes[i][3]]:Hide()
            cql[modes[i][1]]:Show()
        end
    end
    cql:Update()
end

-- debug function to print where a function was called from
function cql:CallerID()
    local where = (debugstack():match(".-\n.-\n.-\n.-\\AddOns\\.-\\(.-:%d+.-)\n") or ""):gsub("\"]","")
    if where:len()==0 then
        where = (debugstack():match(".-\n.-\n.-\\AddOns\\.-\\(.-:%d+.-)\n") or ""):gsub("\"]","")
    end
    print("CallerID:",where:len()>0 and where or debugstack())
end

--[[ binding override ]]

-- rather than hook ToggleQuestLog() and then dealing with the default's stuff being summoned and then
-- having to hide it and stuff, this will override the TOGGLEQUESTLOG bind to run cql:ToggleWindow() instead
local toggleButton = CreateFrame("Button","ClassicQuestLogToggleButton",nil,"SecureActionButtonTemplate")
toggleButton:SetAttribute("type","macro")
toggleButton:SetAttribute("macrotext","/run ClassicQuestLog:ToggleWindow()")
toggleButton:RegisterForClicks("AnyUp","AnyDown")

-- if we are leaving combat, update bind that was waiting while we were in combat
toggleButton:SetScript("OnEvent",function(self,event,...)
    self:UnregisterEvent("PLAYER_REGEN_ENABLED")
    cql:DelayedUpdateBindings()
end)

-- UPDATE_BINDINGS can commonly trigger in sets, and may nil the bind in one event and set it in another.
-- this will wait 0.5 seconds after a set of UPDATE_BINDINGs to run DelayedUpdateBindings
function cql:UPDATE_BINDINGS(...)
    if toggleButton.timer then
        toggleButton.timer:Cancel()
    end
    toggleButton.timer = C_Timer.NewTimer(0.5,cql.DelayedUpdateBindings)
end

-- runs 0.5 seconds after the last UPDATE_BINDING or upon PLAYER_REGEN_ENABLED (leaving combat)
function cql:DelayedUpdateBindings()
    if InCombatLockdown() then -- if we landed here in combat (when we can't change binds), come back after combat
        toggleButton:RegisterEvent("PLAYER_REGEN_ENABLED")
    else
        local key = GetBindingKey("TOGGLEQUESTLOG")
        if toggleButton.key ~= key then -- if TOGGLEQUESTLOG bind changed, set override bind to new bind
            ClearOverrideBindings(toggleButton)
            if key and not cql.options:Get("DontOverrideBind") then
                SetOverrideBindingClick(toggleButton,true,key,"ClassicQuestLogToggleButton","LeftButton")
            end
        end
        toggleButton.key = key
    end
end

-- if DontOverrideBind is enabled, then disable watching for binds changing
function cql:UpdateBindingOverride()
    toggleButton.key = nil
    if cql.options:Get("DontOverrideBind") then
        cql:UnregisterEvent("UPDATE_BINDINGS")
        ClearOverrideBindings(toggleButton)
    else
        cql:RegisterEvent("UPDATE_BINDINGS")
        cql:UPDATE_BINDINGS()
    end
    ClassicQuestLogMicroButton:SetShown(not cql.options:Get("DontOverrideBind"))
end
